// Module level logging for Go
//
// This package provides an alternative to the standard library log package.
//
// The actual logging functions never return errors.  If you are logging
// something, you really don't want to be worried about the actual logging
// having trouble.
//
// Modules have names that are defined by dotted strings.
//   e.g.  "first.second.third"
//
// There is a root module that is reflected by the name "".  Each module
// (except the root module) has a parent, that is identified by the part of
// the name without the last dotted value.
//   e.g. the parent of "first.second.third" is "first.second"
//        the parent of "first.second" is "first"
//        the parent of "first" is "" (the root module)
//
// Each module can specify its own severity level.  Logging calls that are of
// a lower severity than the module's effective severity level are not written
// out.
//
// Loggers are created using the GetLogger function.
//   e.g. logger := loggo.GetLogger("foo.bar")
//
// By default there is one writer registered, which will write to Stderr, and
// only emit warnings and above.  If you want to continue using the default
// logger, but have it emit all logging levels you need to do the following.
//
//   writer, _, err := loggo.RemoveWriter("default")
//   // err is non-nil if and only if the name isn't found.
//   loggo.RegisterWriter("default", writer, loggo.TRACE)
package loggo

import (
	"fmt"
	"runtime"
	"sort"
	"strings"
	"sync"
	"time"
)

type Level int

const (
	NOT_SPECIFIED Level = iota
	TRACE
	DEBUG
	INFO
	WARNING
	ERROR
	CRITICAL
)

// A Logger with a null impl pointer is effectively a root logger.
type Logger struct {
	impl *module
}

type module struct {
	name   string
	level  Level
	parent *module
}

// Initially the modules map only contains the root module.
var (
	root         = &module{"", WARNING, nil}
	modulesMutex sync.Mutex
	modules      = map[string]*module{
		"": root,
	}
)

func (level Level) String() string {
	switch level {
	case NOT_SPECIFIED:
		return "NOT_SPECIFIED"
	case TRACE:
		return "TRACE"
	case DEBUG:
		return "DEBUG"
	case INFO:
		return "INFO"
	case WARNING:
		return "WARNING"
	case ERROR:
		return "ERROR"
	case CRITICAL:
		return "CRITICAL"
	}
	return "<unknown>"
}

// getLoggerInternal assumes that the modulesMutex is locked.
func getLoggerInternal(name string) Logger {
	impl, found := modules[name]
	if found {
		return Logger{impl}
	}

	parent := root
	// Really wanting rsplit.
	parts := strings.Split(name, ".")
	if len(parts) > 1 {
		parent = getLoggerInternal(strings.Join(parts[:len(parts)-1], ".")).impl
	}
	impl = &module{name, NOT_SPECIFIED, parent}
	modules[name] = impl
	return Logger{impl}
}

func GetLogger(name string) Logger {
	// Lowercase the module name, and look for it in the modules map.
	name = strings.ToLower(name)
	modulesMutex.Lock()
	defer modulesMutex.Unlock()
	return getLoggerInternal(name)
}

func DumpLoggingLevels() string {
	output := []string{}
	// output in alphabetical order.
	keys := []string{}
	modulesMutex.Lock()
	defer modulesMutex.Unlock()
	for key := range modules {
		keys = append(keys, key)
	}
	sort.Strings(keys)
	for _, name := range keys {
		mod := modules[name]
		severity := mod.level
		if severity == NOT_SPECIFIED {
			continue
		}
		module_name := mod.name
		if module_name == "" {
			module_name = "<root>"
		}
		output = append(output, fmt.Sprintf("%s %s", module_name, severity))
	}
	return strings.Join(output, "\n")
}

func ConfigureLogging(configString string) {
	if configString == "" {
		return
	}

	values := strings.FieldsFunc(configString, func(r rune) bool { return r == ';' || r == ':' })
	for _, value := range values {
		s := strings.SplitN(value, "=", 2)
		if len(s) < 2 {
			continue
		}
		name := s[0]
		level := s[1]
		if name == "<root>" {
			name = ""
		}
		logger := GetLogger(name)
		logger.SetLogLevel(GetLoggingLevel(level))
	}
}

// GetLoggingLevel converts a string representation of a logging level to the
// enumerated value.
func GetLoggingLevel(level string) Level {
	level = strings.ToUpper(level)
	switch level {
	case "TRACE":
		return TRACE
	case "DEBUG":
		return DEBUG
	case "INFO":
		return INFO
	case "WARN", "WARNING":
		return WARNING
	case "ERROR":
		return ERROR
	}
	return WARNING
}

func (logger *Logger) getModule() *module {
	if logger == nil || logger.impl == nil {
		return root
	}
	return logger.impl
}

func (logger *Logger) Module() string {
	return logger.getModule().name
}

func (logger *Logger) GetLogLevel() Level {
	return logger.getModule().level
}

func (module *module) getEffectiveLogLevel() Level {
	if module.level == NOT_SPECIFIED && module.parent != nil {
		return module.parent.getEffectiveLogLevel()
	}
	return module.level
}

func (logger *Logger) GetEffectiveLogLevel() Level {
	return logger.getModule().getEffectiveLogLevel()
}

func (logger *Logger) SetLogLevel(level Level) {
	module := logger.getModule()
	// The root module can't be unspecified.
	if module.name == "" && level == NOT_SPECIFIED {
		level = WARNING
	}
	module.level = level
}

func (logger *Logger) Log(level Level, message string, args ...interface{}) {
	if logger.getModule().getEffectiveLogLevel() > level ||
		level < TRACE ||
		level > CRITICAL {
		return
	}
	// Gather time, and filename, line number.
	now := time.Now() // get this early.
	var file string
	var line int
	// Param to Caller is the call depth.  Since this method is called from
	// the Logger methods, we want the place that those were called from.
	_, file, line, ok := runtime.Caller(2)
	if !ok {
		file = "???"
		line = 0
	}

	formattedMessage := fmt.Sprintf(message, args...)
	writeToWriters(level, logger.impl.name, file, line, now, formattedMessage)
}

func (logger *Logger) Critical(message string, args ...interface{}) {
	logger.Log(CRITICAL, message, args...)
}

func (logger *Logger) Error(message string, args ...interface{}) {
	logger.Log(ERROR, message, args...)
}

func (logger *Logger) Warning(message string, args ...interface{}) {
	logger.Log(WARNING, message, args...)
}

func (logger *Logger) Info(message string, args ...interface{}) {
	logger.Log(INFO, message, args...)
}

func (logger *Logger) Debug(message string, args ...interface{}) {
	logger.Log(DEBUG, message, args...)
}

func (logger *Logger) Trace(message string, args ...interface{}) {
	logger.Log(TRACE, message, args...)
}

func (logger *Logger) IsErrorEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= ERROR
}

func (logger *Logger) IsWarningEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= WARNING
}

func (logger *Logger) IsInfoEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= INFO
}

func (logger *Logger) IsDebugEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= DEBUG
}

func (logger *Logger) IsTraceEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= TRACE
}
