package loggo_test

import (
	"fmt"
	"strings"
	"testing"

	. "launchpad.net/gocheck"
	"launchpad.net/loggo"
)

func Test(t *testing.T) {
	TestingT(t)
}

type loggerSuite struct{}

var _ = Suite(&loggerSuite{})

func (*loggerSuite) SetUpTest(c *C) {
	loggo.ResetLogging()
}

func (*loggerSuite) TestRootLogger(c *C) {
	root := loggo.Logger{}
	c.Assert(root.Module(), Equals, "")
	c.Assert(root.IsErrorEnabled(), Equals, true)
	c.Assert(root.IsWarningEnabled(), Equals, true)
	c.Assert(root.IsInfoEnabled(), Equals, false)
	c.Assert(root.IsDebugEnabled(), Equals, false)
	c.Assert(root.IsTraceEnabled(), Equals, false)
}

func (*loggerSuite) TestNullLoggerPointerAsRoot(c *C) {
	var root *loggo.Logger = nil
	c.Assert(root.Module(), Equals, "")
	c.Assert(root.IsErrorEnabled(), Equals, true)
	c.Assert(root.IsWarningEnabled(), Equals, true)
	c.Assert(root.IsInfoEnabled(), Equals, false)
	c.Assert(root.IsDebugEnabled(), Equals, false)
	c.Assert(root.IsTraceEnabled(), Equals, false)
}

func (*loggerSuite) TestModuleName(c *C) {
	logger := loggo.GetLogger("loggo.testing")
	c.Assert(logger.Module(), Equals, "loggo.testing")
}

func (*loggerSuite) TestSetLevel(c *C) {
	logger := loggo.GetLogger("testing")

	c.Assert(logger.GetLogLevel(), Equals, loggo.NOT_SPECIFIED)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.WARNING)
	c.Assert(logger.IsErrorEnabled(), Equals, true)
	c.Assert(logger.IsWarningEnabled(), Equals, true)
	c.Assert(logger.IsInfoEnabled(), Equals, false)
	c.Assert(logger.IsDebugEnabled(), Equals, false)
	c.Assert(logger.IsTraceEnabled(), Equals, false)
	logger.SetLogLevel(loggo.TRACE)
	c.Assert(logger.GetLogLevel(), Equals, loggo.TRACE)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.TRACE)
	c.Assert(logger.IsErrorEnabled(), Equals, true)
	c.Assert(logger.IsWarningEnabled(), Equals, true)
	c.Assert(logger.IsInfoEnabled(), Equals, true)
	c.Assert(logger.IsDebugEnabled(), Equals, true)
	c.Assert(logger.IsTraceEnabled(), Equals, true)
	logger.SetLogLevel(loggo.DEBUG)
	c.Assert(logger.GetLogLevel(), Equals, loggo.DEBUG)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.DEBUG)
	c.Assert(logger.IsErrorEnabled(), Equals, true)
	c.Assert(logger.IsWarningEnabled(), Equals, true)
	c.Assert(logger.IsInfoEnabled(), Equals, true)
	c.Assert(logger.IsDebugEnabled(), Equals, true)
	c.Assert(logger.IsTraceEnabled(), Equals, false)
	logger.SetLogLevel(loggo.INFO)
	c.Assert(logger.GetLogLevel(), Equals, loggo.INFO)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.INFO)
	c.Assert(logger.IsErrorEnabled(), Equals, true)
	c.Assert(logger.IsWarningEnabled(), Equals, true)
	c.Assert(logger.IsInfoEnabled(), Equals, true)
	c.Assert(logger.IsDebugEnabled(), Equals, false)
	c.Assert(logger.IsTraceEnabled(), Equals, false)
	logger.SetLogLevel(loggo.WARNING)
	c.Assert(logger.GetLogLevel(), Equals, loggo.WARNING)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.WARNING)
	c.Assert(logger.IsErrorEnabled(), Equals, true)
	c.Assert(logger.IsWarningEnabled(), Equals, true)
	c.Assert(logger.IsInfoEnabled(), Equals, false)
	c.Assert(logger.IsDebugEnabled(), Equals, false)
	c.Assert(logger.IsTraceEnabled(), Equals, false)
	logger.SetLogLevel(loggo.ERROR)
	c.Assert(logger.GetLogLevel(), Equals, loggo.ERROR)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(logger.IsErrorEnabled(), Equals, true)
	c.Assert(logger.IsWarningEnabled(), Equals, false)
	c.Assert(logger.IsInfoEnabled(), Equals, false)
	c.Assert(logger.IsDebugEnabled(), Equals, false)
	c.Assert(logger.IsTraceEnabled(), Equals, false)
	// This is added for completeness, but not really expected to be used.
	logger.SetLogLevel(loggo.CRITICAL)
	c.Assert(logger.GetLogLevel(), Equals, loggo.CRITICAL)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.CRITICAL)
	c.Assert(logger.IsErrorEnabled(), Equals, false)
	c.Assert(logger.IsWarningEnabled(), Equals, false)
	c.Assert(logger.IsInfoEnabled(), Equals, false)
	c.Assert(logger.IsDebugEnabled(), Equals, false)
	c.Assert(logger.IsTraceEnabled(), Equals, false)
	logger.SetLogLevel(loggo.NOT_SPECIFIED)
	c.Assert(logger.GetLogLevel(), Equals, loggo.NOT_SPECIFIED)
	c.Assert(logger.GetEffectiveLogLevel(), Equals, loggo.WARNING)
}

func (*loggerSuite) TestLevelsSharedForSameModule(c *C) {
	logger1 := loggo.GetLogger("testing.module")
	logger2 := loggo.GetLogger("testing.module")

	logger1.SetLogLevel(loggo.INFO)
	c.Assert(logger1.IsInfoEnabled(), Equals, true)
	c.Assert(logger2.IsInfoEnabled(), Equals, true)
}

func (*loggerSuite) TestModuleLowered(c *C) {
	logger1 := loggo.GetLogger("TESTING.MODULE")
	logger2 := loggo.GetLogger("Testing")

	c.Assert(logger1.Module(), Equals, "testing.module")
	c.Assert(logger2.Module(), Equals, "testing")
}

func (*loggerSuite) TestLevelsInherited(c *C) {
	root := loggo.GetLogger("")
	first := loggo.GetLogger("first")
	second := loggo.GetLogger("first.second")

	root.SetLogLevel(loggo.ERROR)
	c.Assert(root.GetLogLevel(), Equals, loggo.ERROR)
	c.Assert(root.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(first.GetLogLevel(), Equals, loggo.NOT_SPECIFIED)
	c.Assert(first.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(second.GetLogLevel(), Equals, loggo.NOT_SPECIFIED)
	c.Assert(second.GetEffectiveLogLevel(), Equals, loggo.ERROR)

	first.SetLogLevel(loggo.DEBUG)
	c.Assert(root.GetLogLevel(), Equals, loggo.ERROR)
	c.Assert(root.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(first.GetLogLevel(), Equals, loggo.DEBUG)
	c.Assert(first.GetEffectiveLogLevel(), Equals, loggo.DEBUG)
	c.Assert(second.GetLogLevel(), Equals, loggo.NOT_SPECIFIED)
	c.Assert(second.GetEffectiveLogLevel(), Equals, loggo.DEBUG)

	second.SetLogLevel(loggo.INFO)
	c.Assert(root.GetLogLevel(), Equals, loggo.ERROR)
	c.Assert(root.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(first.GetLogLevel(), Equals, loggo.DEBUG)
	c.Assert(first.GetEffectiveLogLevel(), Equals, loggo.DEBUG)
	c.Assert(second.GetLogLevel(), Equals, loggo.INFO)
	c.Assert(second.GetEffectiveLogLevel(), Equals, loggo.INFO)

	first.SetLogLevel(loggo.NOT_SPECIFIED)
	c.Assert(root.GetLogLevel(), Equals, loggo.ERROR)
	c.Assert(root.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(first.GetLogLevel(), Equals, loggo.NOT_SPECIFIED)
	c.Assert(first.GetEffectiveLogLevel(), Equals, loggo.ERROR)
	c.Assert(second.GetLogLevel(), Equals, loggo.INFO)
	c.Assert(second.GetEffectiveLogLevel(), Equals, loggo.INFO)
}

func (*loggerSuite) TestGetLoggingLevel(c *C) {
	c.Assert(loggo.GetLoggingLevel("trace"), Equals, loggo.TRACE)
	c.Assert(loggo.GetLoggingLevel("TrAce"), Equals, loggo.TRACE)
	c.Assert(loggo.GetLoggingLevel("TRACE"), Equals, loggo.TRACE)
	c.Assert(loggo.GetLoggingLevel("debug"), Equals, loggo.DEBUG)
	c.Assert(loggo.GetLoggingLevel("DEBUG"), Equals, loggo.DEBUG)
	c.Assert(loggo.GetLoggingLevel("info"), Equals, loggo.INFO)
	c.Assert(loggo.GetLoggingLevel("INFO"), Equals, loggo.INFO)
	c.Assert(loggo.GetLoggingLevel("warn"), Equals, loggo.WARNING)
	c.Assert(loggo.GetLoggingLevel("WARN"), Equals, loggo.WARNING)
	c.Assert(loggo.GetLoggingLevel("warning"), Equals, loggo.WARNING)
	c.Assert(loggo.GetLoggingLevel("WARNING"), Equals, loggo.WARNING)
	c.Assert(loggo.GetLoggingLevel("error"), Equals, loggo.ERROR)
	c.Assert(loggo.GetLoggingLevel("ERROR"), Equals, loggo.ERROR)
	// Unknown levels result in WARNING
	c.Assert(loggo.GetLoggingLevel("critical"), Equals, loggo.WARNING)
	c.Assert(loggo.GetLoggingLevel("not_specified"), Equals, loggo.WARNING)
	c.Assert(loggo.GetLoggingLevel("other"), Equals, loggo.WARNING)
}

func (*loggerSuite) TestLevelStringValue(c *C) {
	c.Assert(fmt.Sprint(loggo.NOT_SPECIFIED), Equals, "NOT_SPECIFIED")
	c.Assert(fmt.Sprint(loggo.TRACE), Equals, "TRACE")
	c.Assert(fmt.Sprint(loggo.DEBUG), Equals, "DEBUG")
	c.Assert(fmt.Sprint(loggo.INFO), Equals, "INFO")
	c.Assert(fmt.Sprint(loggo.WARNING), Equals, "WARNING")
	c.Assert(fmt.Sprint(loggo.ERROR), Equals, "ERROR")
	c.Assert(fmt.Sprint(loggo.CRITICAL), Equals, "CRITICAL")
	// other values are unknown
	c.Assert(fmt.Sprint(loggo.Level(-1)), Equals, "<unknown>")
	c.Assert(fmt.Sprint(loggo.Level(42)), Equals, "<unknown>")
}

func (*loggerSuite) TestConfigureLoggingEmpty(c *C) {
	root := loggo.GetLogger("")
	root.SetLogLevel(loggo.DEBUG)
	logger := loggo.GetLogger("test.module")
	logger.SetLogLevel(loggo.INFO)
	// Configure passed a null pointer does nothing.
	loggo.ConfigureLogging("")
	expected := `
<root> DEBUG
test.module INFO
`
	c.Assert(loggo.DumpLoggingLevels(), Equals, strings.TrimSpace(expected))
}

func (*loggerSuite) TestConfigureLoggingRoot(c *C) {
	loggo.ConfigureLogging("<root>=debug")
	c.Assert(loggo.DumpLoggingLevels(), Equals, "<root> DEBUG")
}

func (*loggerSuite) TestConfigureLoggingSingleModule(c *C) {
	loggo.ConfigureLogging("test.module=debug")
	expected := `
<root> WARNING
test.module DEBUG
`
	c.Assert(loggo.DumpLoggingLevels(), Equals, strings.TrimSpace(expected))
}

func (*loggerSuite) TestConfigureLoggingMultipleModules(c *C) {
	loggo.ConfigureLogging("module=info;sub.module=debug:other.module=warning")
	expected := `
<root> WARNING
module INFO
other.module WARNING
sub.module DEBUG
`
	c.Assert(loggo.DumpLoggingLevels(), Equals, strings.TrimSpace(expected))
}

type logwriterSuite struct {
	logger loggo.Logger
	writer *testWriter
}

var _ = Suite(&logwriterSuite{})

func (s *logwriterSuite) SetUpTest(c *C) {
	loggo.ResetLogging()
	loggo.RemoveWriter("default")
	s.writer = &testWriter{}
	err := loggo.RegisterWriter("test", s.writer, loggo.TRACE)
	c.Assert(err, IsNil)
	s.logger = loggo.GetLogger("test.writer")
	// Make it so the logger itself writes all messages.
	s.logger.SetLogLevel(loggo.TRACE)
}

func (s *logwriterSuite) TearDownTest(c *C) {
	loggo.ResetWriters()
}

func (s *logwriterSuite) TestLogDoesntLogWeirdLevels(c *C) {
	s.logger.Log(loggo.NOT_SPECIFIED, "message")
	c.Assert(s.writer.log, HasLen, 0)

	s.logger.Log(loggo.Level(42), "message")
	c.Assert(s.writer.log, HasLen, 0)

	s.logger.Log(loggo.CRITICAL+loggo.Level(1), "message")
	c.Assert(s.writer.log, HasLen, 0)

	s.logger.Log(loggo.Level(-1), "message")
	c.Assert(s.writer.log, HasLen, 0)
}
