package loggo

import (
	"fmt"
	"io"
	"os"
	"sync"
	"time"
)

type Writer interface {
	Write(level Level, module, filename string, line int, timestamp time.Time, message string)
}

type registeredWriter struct {
	writer Writer
	level  Level
}

type SimpleWriter struct {
	writer    io.Writer
	formatter Formatter
}

// NewSimpleWriter creates a new simple writer with the specified log formatter.
func NewSimpleWriter(writer io.Writer, formatter Formatter) Writer {
	return &SimpleWriter{writer, formatter}
}

// The registered writers is an application level singleton.
const defaultName = "default"

var (
	writerMutex sync.Mutex
	writers     = map[string]*registeredWriter{
		defaultName: &registeredWriter{
			writer: NewSimpleWriter(os.Stderr, &DefaultFormatter{}),
			level:  TRACE,
		},
	}
)

// ReplaceDefaultWriter is a convenience method that does the equivalent of
// RemoveWriter and then RegisterWriter with the name "default".  The existing
// default writer is returned.
func ReplaceDefaultWriter(writer Writer) (Writer, error) {
	if writer == nil {
		return nil, fmt.Errorf("Writer cannot be nil")
	}
	writerMutex.Lock()
	defer writerMutex.Unlock()
	reg, found := writers[defaultName]
	if !found {
		return nil, fmt.Errorf("There no %q writer", defaultName)
	}
	oldWriter := reg.writer
	reg.writer = writer
	return oldWriter, nil

}

// RegisterWriter adds the writer to the list of writers that get notified
// when logging.  When registering, the caller specifies the minimum logging
// level that will be written, and a name for the writer.  If there is already
// a registered writer with that name, and error is returned.
func RegisterWriter(name string, writer Writer, minLevel Level) error {
	if writer == nil {
		return fmt.Errorf("Writer cannot be nil")
	}
	writerMutex.Lock()
	defer writerMutex.Unlock()
	if _, found := writers[name]; found {
		return fmt.Errorf("There is already a Writer registerd with the name %q", name)
	}
	writers[name] = &registeredWriter{writer: writer, level: minLevel}
	return nil
}

// RemoveWriter removes the Writer identified by 'name' and returns it.
// If the Writer is not found, an error is returned.
func RemoveWriter(name string) (Writer, Level, error) {
	writerMutex.Lock()
	defer writerMutex.Unlock()
	registered, found := writers[name]
	if !found {
		return nil, NOT_SPECIFIED, fmt.Errorf("No registered Writer called %q", name)
	}
	delete(writers, name)
	return registered.writer, registered.level, nil
}

func writeToWriters(level Level, module, filename string, line int, timestamp time.Time, message string) {
	writerMutex.Lock()
	defer writerMutex.Unlock()
	for _, registered := range writers {
		if level >= registered.level {
			registered.writer.Write(level, module, filename, line, timestamp, message)
		}
	}
}

func (simple *SimpleWriter) Write(level Level, module, filename string, line int, timestamp time.Time, message string) {
	logLine := simple.formatter.Format(level, module, filename, line, timestamp, message)
	fmt.Fprintln(simple.writer, logLine)
}
