package loggo_test

import (
	"fmt"
	"time"

	. "launchpad.net/gocheck"
	"launchpad.net/loggo"
)

type writerBasicsSuite struct{}

var _ = Suite(&writerBasicsSuite{})

func (s *writerBasicsSuite) TearDownTest(c *C) {
	loggo.ResetWriters()
}

func (*writerBasicsSuite) TestRemoveDefaultWriter(c *C) {
	defaultWriter, level, err := loggo.RemoveWriter("default")
	c.Assert(err, IsNil)
	c.Assert(level, Equals, loggo.TRACE)
	c.Assert(defaultWriter, Not(IsNil))

	// Trying again fails.
	defaultWriter, level, err = loggo.RemoveWriter("default")
	c.Assert(err, ErrorMatches, `No registered Writer called "default"`)
	c.Assert(level, Equals, loggo.NOT_SPECIFIED)
	c.Assert(defaultWriter, IsNil)
}

func (*writerBasicsSuite) TestRegisterWriterExistingName(c *C) {
	err := loggo.RegisterWriter("default", &testWriter{}, loggo.INFO)
	c.Assert(err, ErrorMatches, `There is already a Writer registerd with the name "default"`)
}

func (*writerBasicsSuite) TestRegisterNilWriter(c *C) {
	err := loggo.RegisterWriter("nil", nil, loggo.INFO)
	c.Assert(err, ErrorMatches, `Writer cannot be nil`)
}

func (*writerBasicsSuite) TestRegisterWriterTypedNil(c *C) {
	// If the interface is a typed nil, we have to trust the user.
	var writer *testWriter
	err := loggo.RegisterWriter("nil", writer, loggo.INFO)
	c.Assert(err, IsNil)
}

type writerSuite struct {
	logger loggo.Logger
}

var _ = Suite(&writerSuite{})

func (s *writerSuite) SetUpTest(c *C) {
	loggo.ResetLogging()
	loggo.RemoveWriter("default")
	s.logger = loggo.GetLogger("test.writer")
	// Make it so the logger itself writes all messages.
	s.logger.SetLogLevel(loggo.TRACE)
}

func (s *writerSuite) TearDownTest(c *C) {
	loggo.ResetWriters()
}

func (s *writerSuite) TearDownSuite(c *C) {
	loggo.ResetLogging()
}

func (s *writerSuite) TestWritingCapturesFileAndLineAndModule(c *C) {
	writer := &testWriter{}
	err := loggo.RegisterWriter("test", writer, loggo.INFO)
	c.Assert(err, IsNil)

	s.logger.Info("Info message")

	// WARNING: test checks the line number of the above logger lines, this
	// will mean that if the above line moves, the test will fail unless
	// updated.
	c.Assert(writer.log, HasLen, 1)
	c.Assert(writer.log[0].filename, Equals, "writer_test.go")
	c.Assert(writer.log[0].line, Equals, 76)
	c.Assert(writer.log[0].module, Equals, "test.writer")
}

func (s *writerSuite) TestWritingLimitWarning(c *C) {
	writer := &testWriter{}
	err := loggo.RegisterWriter("test", writer, loggo.WARNING)
	c.Assert(err, IsNil)

	start := time.Now()
	s.logger.Critical("Something critical.")
	s.logger.Error("An error.")
	s.logger.Warning("A warning message")
	s.logger.Info("Info message")
	s.logger.Trace("Trace the function")
	end := time.Now()

	c.Assert(writer.log, HasLen, 3)
	c.Assert(writer.log[0].level, Equals, loggo.CRITICAL)
	c.Assert(writer.log[0].message, Equals, "Something critical.")
	c.Assert(writer.log[0].timestamp, Between(start, end))

	c.Assert(writer.log[1].level, Equals, loggo.ERROR)
	c.Assert(writer.log[1].message, Equals, "An error.")
	c.Assert(writer.log[1].timestamp, Between(start, end))

	c.Assert(writer.log[2].level, Equals, loggo.WARNING)
	c.Assert(writer.log[2].message, Equals, "A warning message")
	c.Assert(writer.log[2].timestamp, Between(start, end))
}

func (s *writerSuite) TestWritingLimitTrace(c *C) {
	writer := &testWriter{}
	err := loggo.RegisterWriter("test", writer, loggo.TRACE)
	c.Assert(err, IsNil)

	start := time.Now()
	s.logger.Critical("Something critical.")
	s.logger.Error("An error.")
	s.logger.Warning("A warning message")
	s.logger.Info("Info message")
	s.logger.Trace("Trace the function")
	end := time.Now()

	c.Assert(writer.log, HasLen, 5)
	c.Assert(writer.log[0].level, Equals, loggo.CRITICAL)
	c.Assert(writer.log[0].message, Equals, "Something critical.")
	c.Assert(writer.log[0].timestamp, Between(start, end))

	c.Assert(writer.log[1].level, Equals, loggo.ERROR)
	c.Assert(writer.log[1].message, Equals, "An error.")
	c.Assert(writer.log[1].timestamp, Between(start, end))

	c.Assert(writer.log[2].level, Equals, loggo.WARNING)
	c.Assert(writer.log[2].message, Equals, "A warning message")
	c.Assert(writer.log[2].timestamp, Between(start, end))

	c.Assert(writer.log[3].level, Equals, loggo.INFO)
	c.Assert(writer.log[3].message, Equals, "Info message")
	c.Assert(writer.log[3].timestamp, Between(start, end))

	c.Assert(writer.log[4].level, Equals, loggo.TRACE)
	c.Assert(writer.log[4].message, Equals, "Trace the function")
	c.Assert(writer.log[4].timestamp, Between(start, end))
}

func (s *writerSuite) TestMultipleWriters(c *C) {
	errorWriter := &testWriter{}
	err := loggo.RegisterWriter("error", errorWriter, loggo.ERROR)
	c.Assert(err, IsNil)
	warningWriter := &testWriter{}
	err = loggo.RegisterWriter("warning", warningWriter, loggo.WARNING)
	c.Assert(err, IsNil)
	infoWriter := &testWriter{}
	err = loggo.RegisterWriter("info", infoWriter, loggo.INFO)
	c.Assert(err, IsNil)
	traceWriter := &testWriter{}
	err = loggo.RegisterWriter("trace", traceWriter, loggo.TRACE)
	c.Assert(err, IsNil)

	s.logger.Error("An error.")
	s.logger.Warning("A warning message")
	s.logger.Info("Info message")
	s.logger.Trace("Trace the function")

	c.Assert(errorWriter.log, HasLen, 1)
	c.Assert(warningWriter.log, HasLen, 2)
	c.Assert(infoWriter.log, HasLen, 3)
	c.Assert(traceWriter.log, HasLen, 4)
}

func Between(start, end time.Time) Checker {
	if end.Before(start) {
		return &betweenChecker{end, start}
	}
	return &betweenChecker{start, end}
}

type betweenChecker struct {
	start, end time.Time
}

func (checker *betweenChecker) Info() *CheckerInfo {
	info := CheckerInfo{
		Name:   "Between",
		Params: []string{"obtained"},
	}
	return &info
}

func (checker *betweenChecker) Check(params []interface{}, names []string) (result bool, error string) {
	when, ok := params[0].(time.Time)
	if !ok {
		return false, "obtained value type must be time.Time"
	}
	if when.Before(checker.start) {
		return false, fmt.Sprintf("obtained value %#v type must before start value of %#v", when, checker.start)
	}
	if when.After(checker.end) {
		return false, fmt.Sprintf("obtained value %#v type must after end value of %#v", when, checker.end)
	}
	return true, ""
}
